/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_matrix_diag_init
 * lis_matrix_diag_check
 * lis_matrix_diag_create
 * lis_matrix_diag_destroy
 * lis_matrix_diag_duplicate
 * lis_matrix_diag_duplicateM
 * lis_matrix_diag_get_range
 * lis_matrix_diag_get_size
 * lis_matrix_diag_set_blocksize
 ************************************************/

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_init"
int lis_matrix_diag_init(LIS_MATRIX_DIAG *D)
{
	LIS_DEBUG_FUNC_IN;

	memset(*D,0,sizeof(struct LIS_MATRIX_DIAG_STRUCT));
	(*D)->bn = 1;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_check"
int lis_matrix_diag_check(LIS_MATRIX_DIAG D, int level)
{
	LIS_DEBUG_FUNC_IN;

	switch( level )
	{
	case LIS_MATRIX_CHECK_NULL:
		if( D==NULL )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"diagonal matrix D is undefined\n");
			return LIS_ERR_ILL_ARG;
		}
		break;
	default:
		if( D==NULL )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"diagonal matrix D is undefined\n");
			return LIS_ERR_ILL_ARG;
		}
		break;
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_create"
int lis_matrix_diag_create(int local_n, int global_n, LIS_Comm comm, LIS_MATRIX_DIAG *D)
{
	int n,gn;
	int nprocs,my_rank;
	int is,ie;
	int i;
	int *ranges;

	LIS_DEBUG_FUNC_IN;

	*D = NULL;

	if( global_n>0 && local_n>global_n )
	{
		LIS_SETERR2(LIS_ERR_ILL_ARG,"local n(=%d) is larger than global n(=%d)\n",local_n,global_n);
		return LIS_ERR_ILL_ARG;
	}
	if( local_n<0 || global_n<0 )
	{
		LIS_SETERR2(LIS_ERR_ILL_ARG,"local n(=%d) or global n(=%d) are less than 0\n",local_n,global_n);
		return LIS_ERR_ILL_ARG;
	}
	if( local_n==0 && global_n==0 )
	{
		LIS_SETERR2(LIS_ERR_ILL_ARG,"local n(=%d) and global n(=%d) are 0\n",local_n,global_n);
		return LIS_ERR_ILL_ARG;
	}

	*D = (LIS_MATRIX_DIAG)lis_malloc( sizeof(struct LIS_MATRIX_DIAG_STRUCT),"lis_matrix_diag_create::D" );
	if( NULL==*D )
	{
		LIS_SETERR_MEM(sizeof(struct LIS_MATRIX_DIAG_STRUCT));
		return LIS_OUT_OF_MEMORY;
	}
	lis_matrix_diag_init(D);

	#ifdef USE_MPI
		MPI_Comm_size(comm,&nprocs);
		MPI_Comm_rank(comm,&my_rank);
		ranges = (int *)lis_malloc( (nprocs+1)*sizeof(int),"lis_matrix_diag_create::ranges" );
		if( ranges==NULL )
		{
			LIS_SETERR_MEM((nprocs+1)*sizeof(int));
			lis_matrix_diag_destroy(*D);
			*D = NULL;
			return LIS_OUT_OF_MEMORY;
		}
	#else
		nprocs  = 1;
		my_rank = 0;
		ranges  = NULL;
	#endif

	#ifdef USE_MPI
		MPI_Allreduce(&local_n,&i,1,MPI_INT,MPI_SUM,comm);
		if( i==0 )
	#else
		if( local_n==0 )
	#endif
	{
		#ifdef USE_MPI
			LIS_GET_ISIE(my_rank,nprocs,global_n,is,ie);
			local_n = ie-is;
			MPI_Allgather(&ie,1,MPI_INT,&ranges[1],1,MPI_INT,comm);
			ranges[0] = 0;
		#else
			local_n = global_n;
			is      = 0;
			ie      = global_n;
		#endif
	}
	else
	{
		#ifdef USE_MPI
			MPI_Allgather(&local_n,1,MPI_INT,&ranges[1],1,MPI_INT,comm);
			ranges[0] = 0;
			for(i=0;i<nprocs;i++)
			{
				ranges[i+1] += ranges[i];
			}
			global_n = ranges[nprocs];
			is       = ranges[my_rank];
			ie       = ranges[my_rank+1];
		#else
			global_n = local_n;
			is       = 0;
			ie       = local_n;
		#endif
	}
	(*D)->ranges      = ranges;

	(*D)->value = (LIS_SCALAR *)lis_malloc( local_n*sizeof(LIS_SCALAR),"lis_matrix_diag_create::D->value" );
	if( NULL==(*D)->value )
	{
		LIS_SETERR_MEM(local_n*sizeof(LIS_SCALAR));
		lis_matrix_diag_destroy(*D);
		*D = NULL;
		return LIS_OUT_OF_MEMORY;
	}

	(*D)->n           = local_n;
	(*D)->nr          = local_n;
	(*D)->gn          = global_n;
	(*D)->np          = local_n;
	(*D)->comm        = comm;
	(*D)->my_rank     = my_rank;
	(*D)->nprocs      = nprocs;
	(*D)->is          = is;
	(*D)->ie          = ie;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_destroy"
int lis_matrix_diag_destroy(LIS_MATRIX_DIAG D)
{
	LIS_DEBUG_FUNC_IN;

	if( D )
	{
		if( D->value ) lis_free( D->value );
		if( D->work ) lis_free( D->work );
		if( D->bns ) lis_free( D->bns );
		if( D->ptr ) lis_free( D->ptr );
		if( D->ranges ) lis_free( D->ranges );
		lis_free(D);
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_duplicate"
int lis_matrix_diag_duplicate(LIS_MATRIX_DIAG Din, LIS_MATRIX_DIAG *Dout)
{
	int n,np,nnz,err;
	int nprocs;
	int i;
	int *ranges,*w_bns,*ptr;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_diag_check(Din,LIS_MATRIX_CHECK_ALL);
	if( err ) return err;

	nprocs = Din->nprocs;
	*Dout = NULL;

	*Dout = (LIS_MATRIX_DIAG)lis_malloc( sizeof(struct LIS_MATRIX_DIAG_STRUCT),"lis_matrix_diag_duplicate::Dout" );
	if( NULL==*Dout )
	{
		LIS_SETERR_MEM(sizeof(struct LIS_MATRIX_DIAG_STRUCT));
		return LIS_OUT_OF_MEMORY;
	}
	lis_matrix_diag_init(Dout);

	#ifdef USE_MPI
		ranges = (int *)lis_malloc( (nprocs+1)*sizeof(int),"lis_matrix_diag_duplicate::ranges" );
		if( ranges==NULL )
		{
			LIS_SETERR_MEM((nprocs+1)*sizeof(int));
			lis_matrix_diag_destroy(*Dout);
			*Dout = NULL;
			return LIS_OUT_OF_MEMORY;
		}
		for(i=0;i<nprocs+1;i++) ranges[i] = Din->ranges[i];
		(*Dout)->ranges      = ranges;
	#else
		(*Dout)->ranges      = NULL;
	#endif

	if( Din->bns==NULL )
	{
		(*Dout)->value = (LIS_SCALAR *)lis_malloc( Din->bn*Din->bn*Din->nr*sizeof(LIS_SCALAR),"lis_matrix_diag_duplicate::Dout->value" );
		if( NULL==(*Dout)->value )
		{
			LIS_SETERR_MEM(Din->bn*Din->bn*Din->nr*sizeof(LIS_SCALAR));
			lis_matrix_diag_destroy(*Dout);
			*Dout = NULL;
			return LIS_OUT_OF_MEMORY;
		}
		(*Dout)->bn = Din->bn;
	}
	else
	{
		w_bns = (int *)lis_malloc( Din->nr*sizeof(int),"lis_matrix_diag_duplicate::w_bns" );
		if( NULL==w_bns )
		{
			LIS_SETERR_MEM(Din->nr*sizeof(int));
			lis_matrix_diag_destroy(*Dout);
			*Dout = NULL;
			return LIS_OUT_OF_MEMORY;
		}
		(*Dout)->bns = w_bns;
		ptr = (int *)lis_malloc( (Din->nr+1)*sizeof(int),"lis_matrix_diag_duplicate::ptr" );
		if( NULL==ptr )
		{
			LIS_SETERR_MEM((Din->nr+1)*sizeof(int));
			lis_matrix_diag_destroy(*Dout);
			*Dout = NULL;
			return LIS_OUT_OF_MEMORY;
		}
		(*Dout)->ptr = ptr;
		nnz = 0;
		(*Dout)->ptr[0] = 0;
		for(i=0;i<Din->nr;i++)
		{
			nnz              += Din->bns[i]*Din->bns[i];
			(*Dout)->bns[i]   = Din->bns[i];
			(*Dout)->ptr[i+1] = Din->ptr[i+1]; 
		}
		(*Dout)->value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR),"lis_matrix_diag_duplicate::Dout->value" );
		if( NULL==(*Dout)->value )
		{
			LIS_SETERR_MEM(nnz*sizeof(LIS_SCALAR));
			lis_matrix_diag_destroy(*Dout);
			*Dout = NULL;
			return LIS_OUT_OF_MEMORY;
		}
	}

	(*Dout)->n           = Din->n;
	(*Dout)->nr          = Din->nr;
	(*Dout)->gn          = Din->gn;
	(*Dout)->np          = Din->np;
	(*Dout)->comm        = Din->comm;
	(*Dout)->my_rank     = Din->my_rank;
	(*Dout)->nprocs      = Din->nprocs;
	(*Dout)->is          = Din->is;
	(*Dout)->ie          = Din->ie;
	(*Dout)->origin      = Din->origin;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_duplicate"
int lis_matrix_diag_duplicateM(LIS_MATRIX Ain, LIS_MATRIX_DIAG *Dout)
{
	int n,np,err;
	int nprocs;
	int i,k;
	int *ranges;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(Ain,LIS_MATRIX_CHECK_ALL);
	if( err ) return err;

	nprocs = Ain->nprocs;
	*Dout = NULL;

	*Dout = (LIS_MATRIX_DIAG)lis_malloc( sizeof(struct LIS_MATRIX_DIAG_STRUCT),"lis_matrix_diag_duplicateM::Dout" );
	if( NULL==*Dout )
	{
		LIS_SETERR_MEM(sizeof(struct LIS_MATRIX_DIAG_STRUCT));
		return LIS_OUT_OF_MEMORY;
	}
	lis_matrix_diag_init(Dout);

	#ifdef USE_MPI
		ranges = (int *)lis_malloc( (nprocs+1)*sizeof(int),"lis_matrix_diag_duplicateM::ranges" );
		if( ranges==NULL )
		{
			LIS_SETERR_MEM((nprocs+1)*sizeof(int));
			lis_matrix_diag_destroy(*Dout);
			*Dout = NULL;
			return LIS_OUT_OF_MEMORY;
		}
		for(i=0;i<nprocs+1;i++) ranges[i] = Ain->ranges[i];
		(*Dout)->ranges      = ranges;
	#else
		(*Dout)->ranges      = NULL;
	#endif

	switch(Ain->matrix_type)
	{
	case LIS_MATRIX_BSR:
	case LIS_MATRIX_BSC:
		k = Ain->nr*Ain->bnr*Ain->bnc;
		(*Dout)->value = (LIS_SCALAR *)lis_malloc( k*sizeof(LIS_SCALAR),"lis_matrix_diag_duplicateM::Dout->value" );
		if( NULL==(*Dout)->value )
		{
			LIS_SETERR_MEM(k*sizeof(LIS_SCALAR));
			lis_matrix_diag_destroy(*Dout);
			*Dout = NULL;
			return LIS_OUT_OF_MEMORY;
		}
		(*Dout)->bn = Ain->bnr;
		(*Dout)->nr = Ain->nr;
		break;
	default:
		(*Dout)->value = (LIS_SCALAR *)lis_malloc( Ain->np*sizeof(LIS_SCALAR),"lis_matrix_diag_duplicateM::Dout->value" );
		if( NULL==(*Dout)->value )
		{
			LIS_SETERR_MEM(Ain->np*sizeof(LIS_SCALAR));
			lis_matrix_diag_destroy(*Dout);
			*Dout = NULL;
			return LIS_OUT_OF_MEMORY;
		}
		(*Dout)->nr          = Ain->n;
		break;
	}

	(*Dout)->n           = Ain->n;
	(*Dout)->gn          = Ain->gn;
	(*Dout)->np          = Ain->np;
	(*Dout)->comm        = Ain->comm;
	(*Dout)->my_rank     = Ain->my_rank;
	(*Dout)->nprocs      = Ain->nprocs;
	(*Dout)->is          = Ain->is;
	(*Dout)->ie          = Ain->ie;
	(*Dout)->origin      = Ain->origin;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_malloc"
int lis_matrix_diag_mallocM(LIS_MATRIX A, LIS_SCALAR **diag)
{
	int err;
	int k;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_ALL);
	if( err ) return err;

	switch(A->matrix_type)
	{
	case LIS_MATRIX_BSR:
		k = A->nr*A->bnr*A->bnc;
		*diag = (LIS_SCALAR *)lis_malloc( k*sizeof(LIS_SCALAR),"lis_matrix_diag_mallocM::diag" );
		if( NULL==*diag )
		{
			LIS_SETERR_MEM(k*sizeof(LIS_SCALAR));
			*diag = NULL;
			return LIS_OUT_OF_MEMORY;
		}
		break;
	default:
		*diag = (LIS_SCALAR *)lis_malloc( A->n*sizeof(LIS_SCALAR),"lis_matrix_diag_mallocM::diag" );
		if( NULL==*diag )
		{
			LIS_SETERR_MEM(A->n*sizeof(LIS_SCALAR));
			*diag = NULL;
			return LIS_OUT_OF_MEMORY;
		}
		break;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_get_range"
int lis_matrix_diag_get_range(LIS_MATRIX_DIAG D, int *is, int *ie)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_diag_check(D,LIS_MATRIX_CHECK_NULL);
	if( err ) return err;

	*is = D->is;
	*ie = D->ie;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_get_size"
int lis_matrix_diag_get_size(LIS_MATRIX_DIAG D, int *local_n, int *global_n)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_diag_check(D,LIS_MATRIX_CHECK_NULL);
	if( err ) return err;

	*local_n  = D->n;
	*global_n = D->gn;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_set_blocksize"
int lis_matrix_diag_set_blocksize(LIS_MATRIX_DIAG D, int bn, int *bns)
{
	int i,n,nnz,nr;
	int *w_bns,*ptr;
	int err;
	LIS_SCALAR *diag;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_diag_check(D,LIS_MATRIX_CHECK_NULL);
	if( err ) return err;

	n = D->n;
	if( bns==NULL )
	{
		nr   = 1 + (n-1)/bn;
		diag = (LIS_SCALAR *)lis_malloc( bn*bn*nr*sizeof(LIS_SCALAR),"lis_matrix_diag_set_blocksize::diag" );
		if( NULL==diag )
		{
			LIS_SETERR_MEM(bn*bn*nr*sizeof(LIS_SCALAR));
			return LIS_OUT_OF_MEMORY;
		}
		lis_free(D->value);
		D->value = diag;
		D->nr    = nr;
		D->bn    = bn;
	}
	else
	{
		if( D->bns==NULL )
		{
			w_bns = (int *)lis_malloc( bn*sizeof(int),"lis_matrix_diag_set_blocksize::w_bns" );
			if( NULL==w_bns )
			{
				LIS_SETERR_MEM(bn*sizeof(int));
				return LIS_OUT_OF_MEMORY;
			}
			D->bns = w_bns;
			ptr = (int *)lis_malloc( (bn+1)*sizeof(int),"lis_matrix_diag_set_blocksize::ptr" );
			if( NULL==ptr )
			{
				LIS_SETERR_MEM((bn+1)*sizeof(int));
				return LIS_OUT_OF_MEMORY;
			}
			D->ptr = ptr;
		}
		nnz       = 0;
		D->ptr[0] = 0;
		for(i=0;i<bn;i++)
		{
			nnz         += bns[i]*bns[i];
			D->ptr[i+1]  = nnz;
			D->bns[i]    = bns[i];
		}
		diag = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR),"lis_matrix_diag_set_blocksize::diag" );
		if( NULL==diag )
		{
			LIS_SETERR_MEM(nnz*sizeof(LIS_SCALAR));
			return LIS_OUT_OF_MEMORY;
		}
		lis_free(D->value);
		D->value = diag;
		D->nr    = bn;
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_copy"
int lis_matrix_diag_copy(LIS_MATRIX_DIAG X, LIS_MATRIX_DIAG Y)
{
	int i,k,n,nr,bn;
	LIS_SCALAR *x,*y;

	LIS_DEBUG_FUNC_IN;

	x  = X->value;
	y  = Y->value;
	n  = X->n;
	nr = X->nr;
	if( n!=Y->n )
	{
		LIS_SETERR(LIS_ERR_ILL_ARG,"length of diagonal matrix X and Y is not equal\n");
		return LIS_ERR_ILL_ARG;
	}
	if( X->bns )
	{
		#pragma omp parallel for private(i)
		for(i=0;i<nr;i++)
		{
			bn = X->bns[i]*X->bns[i];
			k  = X->ptr[i];
			memcpy(&y[k],&x[k],bn*sizeof(LIS_SCALAR));
		}
	}
	else
	{
		bn = X->bn*X->bn;
		#pragma omp parallel for private(i)
		for(i=0; i<nr; i++)
		{
			memcpy(&y[i*bn],&x[i*bn],bn*sizeof(LIS_SCALAR));
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_scale"
int lis_matrix_diag_scale(LIS_SCALAR alpha, LIS_MATRIX_DIAG D)
{
	int i,j,k,nr,bn;
	LIS_SCALAR *d;

	LIS_DEBUG_FUNC_IN;

	d  = D->value;
	nr = D->nr;
	bn = D->bn;
	if( D->bns )
	{
		#pragma omp parallel for private(i)
		for(i=0; i<nr; i++)
		{
			bn = D->bns[i]*D->bns[i];
			k  = D->ptr[i];
			for(j=0;j<bn;j++)
			{
				d[k+j] = alpha * d[k+j];
			}
		}
	}
	else
	{
		switch(bn)
		{
		case 1:
			#pragma omp parallel for private(i)
			for(i=0; i<nr; i++)
			{
				d[i] = alpha * d[i];
			}
			break;
		case 2:
			#pragma omp parallel for private(i)
			for(i=0; i<nr; i++)
			{
				d[4*i]   = alpha * d[4*i];
				d[4*i+1] = alpha * d[4*i+1];
				d[4*i+2] = alpha * d[4*i+2];
				d[4*i+3] = alpha * d[4*i+3];
			}
			break;
		case 3:
			#pragma omp parallel for private(i)
			for(i=0; i<nr; i++)
			{
				d[9*i]   = alpha * d[9*i];
				d[9*i+1] = alpha * d[9*i+1];
				d[9*i+2] = alpha * d[9*i+2];
				d[9*i+3] = alpha * d[9*i+3];
				d[9*i+4] = alpha * d[9*i+4];
				d[9*i+5] = alpha * d[9*i+5];
				d[9*i+6] = alpha * d[9*i+6];
				d[9*i+7] = alpha * d[9*i+7];
				d[9*i+8] = alpha * d[9*i+8];
			}
			break;
		case 4:
			#pragma omp parallel for private(i)
			for(i=0; i<nr; i++)
			{
				d[16*i]    = alpha * d[16*i];
				d[16*i+1]  = alpha * d[16*i+1];
				d[16*i+2]  = alpha * d[16*i+2];
				d[16*i+3]  = alpha * d[16*i+3];
				d[16*i+4]  = alpha * d[16*i+4];
				d[16*i+5]  = alpha * d[16*i+5];
				d[16*i+6]  = alpha * d[16*i+6];
				d[16*i+7]  = alpha * d[16*i+7];
				d[16*i+8]  = alpha * d[16*i+8];
				d[16*i+9]  = alpha * d[16*i+9];
				d[16*i+10] = alpha * d[16*i+10];
				d[16*i+11] = alpha * d[16*i+11];
				d[16*i+12] = alpha * d[16*i+12];
				d[16*i+13] = alpha * d[16*i+13];
				d[16*i+14] = alpha * d[16*i+14];
				d[16*i+15] = alpha * d[16*i+15];
			}
			break;
		default:
			bn = bn*bn;
			#pragma omp parallel for private(i)
			for(i=0; i<nr; i++)
			{
				for(j=0;j<bn;j++)
				{
					d[i*bn+j] = alpha * d[i*bn+j];
				}
			}
			break;
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_inverse"
int lis_matrix_diag_inverse(LIS_MATRIX_DIAG D)
{
	int i,j,nr,bn,nn,ii,jj,kk;
	LIS_SCALAR t,a,b[9];
	LIS_SCALAR *d,*c;

	LIS_DEBUG_FUNC_IN;

	d  = D->value;
	nr = D->nr;
	bn = D->bn;
	nn = D->bn*D->bn;
	if( D->bns )
	{
		#pragma omp parallel for private(i)
		for(i=0; i<nr; i++)
		{
			bn = D->bns[i];
			for(j=0;j<bn;j++)
			{
			}
		}
	}
	else
	{
		switch(bn)
		{
		case 1:
			#pragma omp parallel for private(i)
			for(i=0; i<nr; i++)
			{
				d[i] = 1.0 / d[i];
			}
			break;
		case 2:
			#pragma omp parallel for private(i)
			for(i=0; i<nr; i++)
			{
				/*
				 D[i] = |d0 d2|
				        |d1 d3|
				 */
				t        = 1.0/(d[4*i]*d[4*i+3] - d[4*i+1]*d[4*i+2]);
				a        = d[4*i];
				d[4*i]   = t*d[4*i+3];
				d[4*i+1] = -t*d[4*i+1];
				d[4*i+2] = -t*d[4*i+2];
				d[4*i+3] = t*a;
			}
			break;
		case 3:
			#pragma omp parallel for private(i)
			for(i=0; i<nr; i++)
			{
				/*
				 D[i] = |d0 d3 d6|
				        |d1 d4 d7|
						|d2 d5 d8|
				 */
				t       = d[9*i]*d[9*i+4]*d[9*i+8] + d[9*i+1]*d[9*i+5]*d[9*i+6] + d[9*i+2]*d[9*i+3]*d[9*i+7];
				t      -= d[9*i]*d[9*i+5]*d[9*i+7] + d[9*i+1]*d[9*i+3]*d[9*i+8] + d[9*i+2]*d[9*i+4]*d[9*i+6];
				t       = 1.0 / t;
				b[0]    = t*(d[9*i+4]*d[9*i+8] - d[9*i+7]*d[9*i+5]);
				b[1]    = t*(d[9*i+7]*d[9*i+2] - d[9*i+1]*d[9*i+8]);
				b[2]    = t*(d[9*i+1]*d[9*i+5] - d[9*i+4]*d[9*i+2]);
				b[3]    = t*(d[9*i+6]*d[9*i+5] - d[9*i+3]*d[9*i+8]);
				b[4]    = t*(d[9*i+0]*d[9*i+8] - d[9*i+6]*d[9*i+2]);
				b[5]    = t*(d[9*i+3]*d[9*i+2] - d[9*i+0]*d[9*i+5]);
				b[6]    = t*(d[9*i+3]*d[9*i+7] - d[9*i+6]*d[9*i+4]);
				b[7]    = t*(d[9*i+6]*d[9*i+1] - d[9*i+0]*d[9*i+7]);
				b[8]    = t*(d[9*i+0]*d[9*i+4] - d[9*i+3]*d[9*i+1]);
				memcpy(d,b,9*sizeof(LIS_SCALAR));
			}
			break;
		default:
			c = (LIS_SCALAR *)lis_malloc(nn*sizeof(LIS_SCALAR),"lis_matrix_diag_inverse::c");
			for(i=0;i<nr;i++)
			{
				/* LU decomposition */
				for(ii=0;ii<bn;ii++)
				{
					for(jj=ii+1;jj<bn;jj++)
					{
						d[i*nn + ii*bn+jj] /= d[i*nn + ii*bn+ii];
						for(kk=ii+1;kk<bn;kk++)
						{
							d[i*nn + kk*bn+jj] -= d[i*nn + kk*bn+ii]*d[i*nn + ii*bn+jj];
						}
					}
				}
				/* compute inverse */
				for(kk=0;kk<bn;kk++)
				{
					for(ii=0;ii<bn;ii++)
					{
						if( kk==ii )
						{
							c[kk*bn+ii] = 1.0;
						}
						else
						{
							c[kk*bn+ii] = 0.0;
						}
					}
					for(ii=0;ii<bn;ii++)
					{
						for(jj=ii+1;jj<bn;jj++)
						{
							c[kk*bn+jj] -= c[kk*bn+ii]*d[i*nn + ii*bn+jj];
						}
					}
					for(ii=bn-1;ii>=0;ii--)
					{
						for(jj=ii+1;jj<bn;jj++)
						{
							c[kk*bn+ii] -= c[kk*bn+jj]*d[i*nn + jj*bn+ii];
						}
						c[kk*bn+ii] /= d[i*nn + ii*bn+ii];
					}
				}
				memcpy(&d[i*nn],c,nn*sizeof(LIS_SCALAR));
			}
			lis_free(c);
			break;
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_diag_print"
int lis_matrix_diag_print(LIS_MATRIX_DIAG D)
{
	int	k,i,j,nr,bn,nn;

	LIS_DEBUG_FUNC_IN;

	nr = D->nr;
	bn = D->bn;
	nn = D->bn*D->bn;
	if( D->bns )
	{
	}
	else
	{
		for(k=0; k<nr; k++)
		{
			for(i=0;i<bn;i++)
			{
				printf("%4d (", k*nn+i);
				for(j=0; j<bn; j++)
				{
					printf("%6.2f ", D->value[k*nn + j*bn+i]);
				}
				printf(")\n");
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
