/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

/*
 * This subroutine is made based on SPLIB.
 *
 * R. Bramley and X. Wang.
 * SPLIB: A library of iterative methods for sparse linear system.
 * Technical report, Indiana University--Bloomington, 1995.
 *
 */
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

int lis_symbolic_fact(LIS_SOLVER solver, LIS_PRECON precon);
int lis_numerical_fact(LIS_SOLVER solver, LIS_PRECON precon);

#undef __FUNC__
#define __FUNC__ "lis_precon_create_iluk"
int lis_precon_create_iluk(LIS_SOLVER solver, LIS_PRECON precon)
{
	int				err;
	LIS_MATRIX		A,B;

	LIS_DEBUG_FUNC_IN;

	switch( solver->A->matrix_type )
	{
	case LIS_MATRIX_CRS:
		err = lis_precon_create_iluk_crs(solver,precon);
		break;
	default:
		A = solver->A;
		err = lis_matrix_duplicate(A,&B);
		if( err ) return err;
		lis_matrix_set_type(B,LIS_MATRIX_CRS);
		err = lis_matrix_convert(A,B);
		if( err ) return err;
		solver->A = B;
		err = lis_precon_create_iluk_crs(solver,precon);
		lis_matrix_destroy(B);
		solver->A = A;
		break;
	}

	precon->is_copy = LIS_TRUE;
	LIS_DEBUG_FUNC_OUT;
    return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_precon_create_iluk_crs"
int lis_precon_create_iluk_crs(LIS_SOLVER solver, LIS_PRECON precon)
{
	int n,nnz,k,err;
	LIS_SCALAR r;
	LIS_MATRIX A,P;
	
	LIS_DEBUG_FUNC_IN;

	A   = solver->A;
	n   = A->n;
	nnz = A->nnz;
	k   = solver->options[LIS_OPTIONS_FILL];
	r   = solver->params[LIS_PARAMS_RELAX-LIS_OPTIONS_LEN];

	if( k<0 ) solver->options[LIS_OPTIONS_FILL] = 0;
	if( r<0.0 || r>1.0 ) solver->params[LIS_PARAMS_RELAX-LIS_OPTIONS_LEN] = 0.0;

	err = lis_matrix_create(A->comm,&P);
	if( err )
	{
		return err;
	}
	err = lis_matrix_set_size(P,n,0);
	if( err )
	{
		return err;
	}

	precon->A = P;
	lis_symbolic_fact(solver,precon);
	lis_numerical_fact(solver,precon);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#define NEW_BLOCK 400
#undef __FUNC__
#define __FUNC__ "lis_symbolic_fact"
int lis_symbolic_fact(LIS_SOLVER solver, LIS_PRECON precon)
{
	int n,nnz,m;
	int levfill,head,lowct,prev,actlev,lm,lmk,ijlum,ujbeg,ujend;
	int i,j,k,jj,rowct,k1,k2,nzi,nzlu,levp1;
	int	*ijlu,*uptr;
	int *lastcol,*levels,*rowll;
	LIS_MATRIX A;

	LIS_DEBUG_FUNC_IN;

	A       = solver->A;
	n       = A->n;
	nnz     = A->nnz;
	levfill = solver->options[LIS_OPTIONS_FILL];

	ijlu    = (int *)lis_malloc( (nnz+1)*sizeof(int),"lis_symbolic_fact::ijlu");  /* Added by Jang */
	uptr    = (int *)lis_malloc( n*sizeof(int),"lis_symbolic_fact::uptr" );
	lastcol = (int *)lis_malloc( n*sizeof(int),"lis_symbolic_fact::lastcol" );
	levels  = (int *)lis_malloc( (nnz+1)*sizeof(int),"lis_symbolic_fact::levels" );
	rowll   = (int *)lis_malloc( n*sizeof(int),"lis_symbolic_fact::rowll" );		/* Added by Jang */

	ijlu[0] = n+1;
	uptr[0] = n+1;

	nzlu    = n;
	levp1   = levfill+1;

	for(i=0;i<n;i++) lastcol[i]=0;

	for(i=0;i<n;i++)
	{
		lis_sort_id(A->ptr[i],A->ptr[i+1]-1,A->index,A->value);
		levels[i] = 0;
		nzi       = A->ptr[i+1]-A->ptr[i];
		if( nzi>1 )
		{
			/* n     = 5
			 * nzi   = 4
			 * rowll = [ 3, 1, 4, -, -] 
			 */
			rowct = 0;
			for(j=A->ptr[i];j<A->ptr[i+1];j++)
			{
				jj          = A->index[j];
				lastcol[jj] = i+1;
				if( jj!=i )
				{
					levels[jj]   = 1;
					rowll[rowct] = jj;
					rowct++;
				}
			}
			/* 
			 * rowll = [ 1, 3, 4, -, -] 
			 */
			lis_sort_i(0,nzi-2,rowll);
			/* 
			 * rowll = [ -, 3, -, 4, 5] 
			 */
			head = rowll[0];
			k1   = n;
			for(j=nzi-2;j>=0;j--)
			{
				k2        = rowll[j];
				rowll[k2] = k1;
				k1        = k2;
			}

			nzlu += nzi-1;
			j = head;
			lowct = 0;
			while( j<i )
			{
				lowct++;
				if( levfill!=0 )
				{
					ujbeg = uptr[j];
					ujend = ijlu[j+1];
					prev  = j;
					lm    = rowll[j];
					lmk   = levels[j];
					for(m=ujbeg;m<ujend;m++)
					{
						ijlum = ijlu[m];
						actlev = lmk + levels[m];
						if( lastcol[ijlum]!=i+1 )
						{
							if( actlev<=levp1 )
							{
								lastcol[ijlum] = i+1;
								levels[ijlum]  = actlev;
								while( lm<=ijlum )
								{
									prev = lm;
									lm   = rowll[lm];
								}
								rowll[prev] = ijlum;
								rowll[ijlum] = lm;
								prev = ijlum;
								nzlu++;
							}
						}
						else
						{
							levels[ijlum] = _min(levels[ijlum],actlev);
						}
					}
				}
				j = rowll[j];
			}
			if( nzlu>=nnz )
			{
				ijlu   = (int *)lis_realloc(ijlu  , (nzlu + NEW_BLOCK)*sizeof(int)); /* Added by Jang. */
				levels = (int *)lis_realloc(levels, (nzlu + NEW_BLOCK)*sizeof(int)); /* Added by Jang. */
				nnz  = nzlu + NEW_BLOCK;
			}
			ijlu[i+1] = nzlu+1;
			uptr[i] = ijlu[i] + lowct;
			j = head;
			k1 = ijlu[i];
			for(k=k1;k<=nzlu;k++)
			{
				ijlu[k] = j;
				levels[k] = levels[j];
				j = rowll[j];
			}
		}
		else
		{
			ijlu[i+1] = nzlu + 1;
			uptr[i] = ijlu[i];
		}
	}
	lis_free(levels);
	lis_free(rowll);
	lis_free(lastcol);
	precon->A->index = ijlu;
	precon->A->ptr   = uptr;
	precon->A->nnz   = nnz;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_numerical_fact"
int lis_numerical_fact(LIS_SOLVER solver, LIS_PRECON precon)
{
	int n,j,k,m,jj,mm,nrw,milu,nnz;
	int *colptrs;
	int *jlu, *uptr;
	LIS_SCALAR *lu;
	LIS_SCALAR rwnrm,mult,relax;
	LIS_MATRIX A;

	LIS_DEBUG_FUNC_IN;

	A       = solver->A;
	jlu     = precon->A->index;
	uptr    = precon->A->ptr;
	nnz     = precon->A->nnz;
	relax   = solver->params[LIS_PARAMS_RELAX-LIS_OPTIONS_LEN];
	n       = A->n;
	milu    = (relax!=0.0);

	colptrs = (int *)lis_malloc( n*sizeof(int),"lis_numerical_fact::colptrs" );						/* Added by Jang. */
	lu      = (LIS_SCALAR *)lis_malloc( (nnz+1)*sizeof(LIS_SCALAR),"lis_numerical_fact::lu" );	/* Added by Jang. */

	for(k=0;k<n;k++) colptrs[k] = 0;

	for(k=0;k<n;k++)
	{
		for(j=jlu[k];j<jlu[k+1];j++)
		{
			colptrs[jlu[j]] = j+1;
			lu[j]           = 0.0;
		}

		colptrs[k] = k+1;
		rwnrm = 0.0;
		nrw   = 0;
		for(j=A->ptr[k];j<A->ptr[k+1];j++)
		{
			jj = A->index[j];
			if( colptrs[jj]!=0 )
			{
				nrw++;
				rwnrm += fabs(A->value[j]);
				lu[colptrs[jj]-1] = A->value[j];
			}
		}

		for(j=jlu[k];j<uptr[k];j++)
		{
			jj = jlu[j];
			lu[j] = lu[j]*lu[jj];
			mult  = lu[j];
			if( milu )
			{
				for(m=uptr[jj];m<jlu[jj+1];m++)
				{
					mm = jlu[m];
					if( colptrs[mm]!=0 )
					{
						lu[colptrs[mm]-1] -= mult*lu[m];
					}
					else
					{
						lu[k] -= relax*mult*lu[m];
					}
				}
			}
			else
			{
				for(m=uptr[jj];m<jlu[jj+1];m++)
				{
					mm = jlu[m];
					if( colptrs[mm]!=0 )
					{
						lu[colptrs[mm]-1] -= mult*lu[m];
					}
				}
			}
		}

		for(j=jlu[k];j<jlu[k+1];j++)
		{
			colptrs[jlu[j]] = 0;
		}
		colptrs[k] = 0;

		lu[k] = 1.0/lu[k];
	}

	lis_free(colptrs);
	precon->A->value = lu;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;

}

#undef __FUNC__
#define __FUNC__ "lis_psolve_iluk"
int lis_psolve_iluk(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,n;
	LIS_SCALAR t;
	LIS_SCALAR *b,*x;
	LIS_PRECON precon;

	LIS_DEBUG_FUNC_IN;
	/*
	 *  LUx = b
	 */

	precon = solver->precon;
	n = precon->A->n;
	b = B->value;
	x = X->value;
	for(i=0; i<n; i++)
	{
		t = b[i];
		for(j=precon->A->index[i];j<precon->A->ptr[i];j++)
		{
			t -= precon->A->value[j] * x[precon->A->index[j]];
		}
		x[i] = t;
	}
	for(i=n-1; i>=0; i--)
	{
		t = x[i];
		for(j=precon->A->ptr[i];j<precon->A->index[i+1];j++)
		{
			t -= precon->A->value[j] * x[precon->A->index[j]];
		}
		x[i] = t * precon->A->value[i];
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolvet_iluk"
int lis_psolvet_iluk(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,n;
	LIS_SCALAR *b,*x;
	LIS_PRECON precon;

	LIS_DEBUG_FUNC_IN;
	/*
	 *  (LU)'x = b
	 */

	precon = solver->precon;
	n = precon->A->n;
	b = B->value;
	x = X->value;
	for(i=0; i<n; i++)
	{
		x[i] = b[i];
	}
	for(i=0; i<n; i++)
	{
		x[i] = x[i] * precon->A->value[i];
		for(j=precon->A->ptr[i];j<precon->A->index[i+1];j++)
		{
			x[precon->A->index[j]] -= precon->A->value[j] * x[i];
		}
	}
	for(i=n-1; i>=0; i--)
	{
		for(j=precon->A->index[i];j<precon->A->ptr[i];j++)
		{
			x[precon->A->index[j]] -= precon->A->value[j] * x[i];
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

