/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_precon_create
 * lis_psolve
 * lis_psolvet
 ************************************************/

#undef __FUNC__
#define __FUNC__ "lis_precon_create_sainv"
int lis_precon_create_sainv(LIS_SOLVER solver, LIS_PRECON precon)
{
	int				err;
	LIS_MATRIX		A,B;


	LIS_DEBUG_FUNC_IN;

	switch( solver->A->matrix_type )
	{
	case LIS_MATRIX_CRS:
		err = lis_precon_create_sainv_crs(solver,precon);
		break;
	default:
		A = solver->A;
		err = lis_matrix_duplicate(A,&B);
		if( err ) return err;
		lis_matrix_set_type(B,LIS_MATRIX_CRS);
		err = lis_matrix_convert(A,B);
		if( err ) return err;
		solver->A = B;
		err = lis_matrix_split(solver->A);
		if( err )
		{
			return err;
		}
		err = lis_precon_create_sainv_crs(solver,precon);
		lis_matrix_destroy(B);
		solver->A = A;
		break;
	}

	#ifndef USE_QUAD_PRECISION
		err = lis_vector_duplicate(solver->A,&precon->temp);
	#else
		if( solver->precision==LIS_PRECISION_DEFAULT )
		{
			err = lis_vector_duplicate(solver->A,&precon->temp);
		}
		else
		{
			err = lis_vector_duplicateex(LIS_PRECISION_QUAD,solver->A,&precon->temp);
		}
	#endif
	if( err ) return err;

	precon->A       = solver->A;
	precon->is_copy = LIS_FALSE;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_precon_create_sainv_crs"
int lis_precon_create_sainv_crs(LIS_SOLVER solver, LIS_PRECON precon)
{
	int				err;
	int				i,j,k,ii,jj,ik,jk;
	int				n,nnz,annz,cl,cu;
	int				*ww,*il,*iu;
	LIS_SCALAR		t,dd,tol;
	LIS_SCALAR		nrm;
	LIS_SCALAR		*d,*l,*u;
	LIS_MATRIX		A,B,W,Z,L,U;
	LIS_VECTOR		D;

	LIS_DEBUG_FUNC_IN;


	A      = solver->A;
	n      = A->n;
	tol    = solver->params[LIS_PARAMS_DROP-LIS_OPTIONS_LEN];
	annz   = A->n / 10;

	B      = NULL;
	W      = NULL;
	Z      = NULL;
	L      = NULL;
	U      = NULL;
	ww     = NULL;
	d      = NULL;
	l      = NULL;
	u      = NULL;
	il     = NULL;
	iu     = NULL;

	err = lis_matrix_merge(A);
	if( err ) return err;
	err = lis_matrix_duplicate(A,&B);
	if( err ) return err;
	err = lis_matrix_convert_crs2ccs(A,B);
	if( err )
	{
		return err;
	}
	err = lis_vector_duplicate(A,&D);
	if( err )
	{
		return err;
	}
	d = D->value;
	err = lis_matrix_create_rco(n,0,A->comm,annz,NULL,&W);
	if( err )
	{
		lis_matrix_destroy(B);
		lis_matrix_destroy(W);
		lis_matrix_destroy(Z);
		lis_free2(6,d,l,u,ww,il,iu);
		return err;
	}
	err = lis_matrix_create_rco(n,0,A->comm,annz,NULL,&Z);
	if( err )
	{
		lis_matrix_destroy(B);
		lis_matrix_destroy(W);
		lis_matrix_destroy(Z);
		lis_free2(6,d,l,u,ww,il,iu);
		return err;
	}
	l   = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR),"lis_precon_create_sainv_crs::l");
	if( l==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(LIS_SCALAR));
		lis_matrix_destroy(B);
		lis_matrix_destroy(W);
		lis_matrix_destroy(Z);
		lis_free2(6,d,l,u,ww,il,iu);
		return LIS_OUT_OF_MEMORY;
	}
	u   = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR),"lis_precon_create_sainv_crs::u");
	if( u==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(LIS_SCALAR));
		lis_matrix_destroy(B);
		lis_matrix_destroy(W);
		lis_matrix_destroy(Z);
		lis_free2(6,d,l,u,ww,il,iu);
		return LIS_OUT_OF_MEMORY;
	}
	il   = (int *)lis_malloc(n*sizeof(int),"lis_precon_create_sainv_crs::il");
	if( il==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(int));
		lis_matrix_destroy(B);
		lis_matrix_destroy(W);
		lis_matrix_destroy(Z);
		lis_free2(6,d,l,u,ww,il,iu);
		return LIS_OUT_OF_MEMORY;
	}
	iu   = (int *)lis_malloc(n*sizeof(int),"lis_precon_create_sainv_crs::iu");
	if( iu==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(int));
		lis_matrix_destroy(B);
		lis_matrix_destroy(W);
		lis_matrix_destroy(Z);
		lis_free2(6,d,l,u,ww,il,iu);
		return LIS_OUT_OF_MEMORY;
	}
	ww   = (int *)lis_malloc(n*sizeof(int),"lis_precon_create_sainv_crs::ww");
	if( ww==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(int));
		lis_matrix_destroy(B);
		lis_matrix_destroy(W);
		lis_matrix_destroy(Z);
		lis_free2(6,d,l,u,ww,il,iu);
		return LIS_OUT_OF_MEMORY;
	}
	

	for(i=0;i<n;i++) ww[i] = 0;
	for(i=0;i<n;i++)
	{
		W->w_value[i][0] = 1.0;
		W->w_index[i][0] = i;
		W->w_row[i]      = 1;
		Z->w_value[i][0] = 1.0;
		Z->w_index[i][0] = i;
		Z->w_row[i]      = 1;
	}

	for(i=0; i<n; i++)
	{
		/* nrm_inf(A[i,:]) */
		nrm = 0.0;
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			nrm = _max(nrm,fabs(A->value[j]));
		}
		nrm = 1.0/nrm;

		/* l = AZ_i */
		cl = 0;
		memset(l,0,n*sizeof(LIS_SCALAR));
		for(k=0;k<Z->w_row[i];k++)
		{
			ii = Z->w_index[i][k];
			for(j=B->ptr[ii];j<B->ptr[ii+1];j++)
			{
				jj     = B->index[j];
				if( jj>i )
				{
					l[jj] += B->value[j]*Z->w_value[i][k];
					if( ww[jj]==0 )
					{
						ww[jj]   = 1;
						il[cl++] = jj;
					}
				}
			}
		}
		for(k=0;k<cl;k++) ww[il[k]] = 0;

		/* u = W_i'A */
		cu = 0;
		memset(u,0,n*sizeof(LIS_SCALAR));
		for(k=0;k<W->w_row[i];k++)
		{
			ii = W->w_index[i][k];
			for(j=A->ptr[ii];j<A->ptr[ii+1];j++)
			{
				jj     = A->index[j];
				u[jj] += A->value[j]*W->w_value[i][k];
				if( jj>i && ww[jj]==0 )
				{
					ww[jj]   = 1;
					iu[cu++] = jj;
				}
			}
		}
		for(k=0;k<cu;k++) ww[iu[k]] = 0;

		/* d_ii = uZ_i */
		t = 0.0;
		for(k=0;k<Z->w_row[i];k++)
		{
			t += u[Z->w_index[i][k]]*Z->w_value[i][k];
		}
		d[i] = 1.0/t;

		/* for j>i, l_j!=0            */
		/* w_j = w_j - (l_j/d_ii)*w_i */
		for(jj=0;jj<cl;jj++)
		{
			j = il[jj];
			dd = l[j]*d[i];
			for(k=0;k<W->w_row[j];k++)
			{
				ww[W->w_index[j][k]] = k+1;
			}
			for(ik=0;ik<W->w_row[i];ik++)
			{
				jk = ww[W->w_index[i][ik]];
				if( jk!=0 )
				{
					t = dd*W->w_value[i][ik];
					if( fabs(t)*nrm > tol )
					{
						W->w_value[j][jk-1] -= t;
					}
					/*
					t = W->w_value[j][jk-1] - dd*W->w_value[i][ik];
					if( fabs(t)*nrm > tol )
					{
						W->w_value[j][jk-1] = t;
					}
					else
					{
						memcpy(&W->w_value[j][jk-1],&W->w_value[j][jk],(W->w_row[j]-jk)*sizeof(LIS_SCALAR));
						memcpy(&W->w_index[j][jk-1],&W->w_index[j][jk],(W->w_row[j]-jk)*sizeof(int));
						W->w_row[j]--;
					}
					*/
				}
				else
				{
					t = dd*W->w_value[i][ik];
					if( fabs(t)*nrm > tol )
					{
						if( W->w_row[j] == W->w_nnz[j] )
						{
							W->w_nnz[j] += W->w_annz;
							err = lis_matrix_realloc_rco(j,W->w_nnz[j],&W->w_index,&W->w_value);
							if( err )
							{
								lis_matrix_destroy(B);
								lis_matrix_destroy(W);
								lis_matrix_destroy(Z);
								lis_free2(6,d,l,u,ww,il,iu);
								return err;
							}
						}
						jk                = W->w_row[j];
						W->w_index[j][jk] = W->w_index[i][ik];
						W->w_value[j][jk] = -t;
						W->w_row[j]++;
					}
				}
			}
			for(k=0;k<W->w_row[j];k++)
			{
				ww[W->w_index[j][k]] = 0;
			}
		}

		/* for j>i, u_j!=0            */
		/* z_j = z_j - (u_j/d_ii)*z_i */
		for(jj=0;jj<cu;jj++)
		{
			j = iu[jj];
			dd = u[j]*d[i];
			for(k=0;k<Z->w_row[j];k++)
			{
				ww[Z->w_index[j][k]] = k+1;
			}
			for(ik=0;ik<Z->w_row[i];ik++)
			{
				jk = ww[Z->w_index[i][ik]];
				if( jk!=0 )
				{
					t = dd*Z->w_value[i][ik];
					if( fabs(t)*nrm > tol )
					{
						Z->w_value[j][jk-1] -= t;
					}
					/*
					t = Z->w_value[j][jk-1] - dd*Z->w_value[i][ik];
					if( fabs(t)*nrm > tol )
					{
						Z->w_value[j][jk-1] = t;
					}
					else
					{
						memcpy(&Z->w_value[j][jk-1],&Z->w_value[j][jk],(Z->w_row[j]-jk)*sizeof(LIS_SCALAR));
						memcpy(&Z->w_index[j][jk-1],&Z->w_index[j][jk],(Z->w_row[j]-jk)*sizeof(int));
						Z->w_row[j]--;
					}
					*/
				}
				else
				{
					t = dd*Z->w_value[i][ik];
					if( fabs(t)*nrm > tol )
					{
						if( Z->w_row[j] == Z->w_nnz[j] )
						{
							Z->w_nnz[j] += Z->w_annz;
							err = lis_matrix_realloc_rco(j,Z->w_nnz[j],&Z->w_index,&Z->w_value);
							if( err )
							{
								lis_matrix_destroy(B);
								lis_matrix_destroy(W);
								lis_matrix_destroy(Z);
								lis_free2(6,d,l,u,ww,il,iu);
								return err;
							}
						}
						jk                = Z->w_row[j];
						Z->w_index[j][jk] = Z->w_index[i][ik];
						Z->w_value[j][jk] = -t;
						Z->w_row[j]++;
					}
				}
			}
			for(k=0;k<Z->w_row[j];k++)
			{
				ww[Z->w_index[j][k]] = 0;
			}
		}
	}

	lis_matrix_destroy(B);
	lis_free2(5,l,u,ww,il,iu);

	W->matrix_type = LIS_MATRIX_RCO;
	W->status      = LIS_MATRIX_RCO;
	err = lis_matrix_duplicate(W,&L);
	lis_matrix_set_type(L,LIS_MATRIX_CRS);
	err = lis_matrix_convert(W,L);
	if( err )
	{
		lis_matrix_destroy(W);
		lis_matrix_destroy(Z);
		return err;
	}
	lis_matrix_destroy(W);
	L->matrix_type = LIS_MATRIX_CCS;
	L->status      = LIS_MATRIX_CCS;

	Z->matrix_type = LIS_MATRIX_RCO;
	Z->status      = LIS_MATRIX_RCO;
	err = lis_matrix_duplicate(Z,&U);
	lis_matrix_set_type(U,LIS_MATRIX_CCS);
	err = lis_matrix_convert(Z,U);
	if( err )
	{
		lis_matrix_destroy(Z);
		return err;
	}
	U->matrix_type = LIS_MATRIX_CRS;
	U->status      = LIS_MATRIX_CRS;
	lis_matrix_destroy(Z);

	precon->L  = L;
	precon->U  = U;
	precon->D  = D;

#if _DEBUG
	printf("W: nnz=%d annz=%d\n",L->nnz,L->nnz/n);
	printf("Z: nnz=%d annz=%d\n",U->nnz,U->nnz/n);
#endif


	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolve_sainv"
int lis_psolve_sainv(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,n,precision;
	LIS_MATRIX		A,W,Z;
	LIS_VECTOR		t,d;
	LIS_PRECON	precon;
	LIS_QUAD_DECLAR;
	#ifdef USE_QUAD_PRECISION
		LIS_QDOUBLE w1,w2;
		LIS_SCALAR *xl;
	#endif

	/*
	 *  x  = Mb
	 *  M  = ZD^{-1}W'
	 */

	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	A = precon->A;
	W = precon->L;
	Z = precon->U;
	d = precon->D;
	t = precon->temp;
	n = precon->L->n;
	#ifdef USE_QUAD_PRECISION
		xl = X->value_lo;
	#endif

	#ifdef USE_QUAD_PRECISION
		if( B->precision==LIS_PRECISION_DEFAULT )
		{
	#endif
			lis_matvect(W,B,X);
			for(i=0;i<n;i++)
			{
				t->value[i] = X->value[i]*d->value[i];
			}
			lis_matvec(Z,t,X);
	#ifdef USE_QUAD_PRECISION
		}
		else
		{
			lis_matvect(W,B,X);
			for(i=0;i<n;i++)
			{
				#ifndef USE_SSE2
					LIS_QUAD_MULD(t->value[i],t->value_lo[i],X->value[i],X->value_lo[i],d->value[i]);
				#else
					LIS_QUAD_MULD_SSE2(t->value[i],t->value_lo[i],X->value[i],X->value_lo[i],d->value[i]);
				#endif
				/* t->value[i] = X->value[i]*d->value[i]; */
			}
			lis_matvec(Z,t,X);
		}
	#endif

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolvet_sainv"
int lis_psolvet_sainv(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,n,precision;
	LIS_MATRIX		A,W,Z;
	LIS_VECTOR		t,d;
	LIS_PRECON	precon;
	LIS_QUAD_DECLAR;
	#ifdef USE_QUAD_PRECISION
		LIS_QDOUBLE w1,w2;
		LIS_SCALAR *xl;
	#endif

	/*
	 *  x  = M'b
	 *  M' = WD^{-1}Z'
	 */

	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	A = precon->A;
	W = precon->L;
	Z = precon->U;
	d = precon->D;
	t = precon->temp;
	n = precon->L->n;
	#ifdef USE_QUAD_PRECISION
		xl = X->value_lo;
	#endif

	#ifdef USE_QUAD_PRECISION
		if( B->precision==LIS_PRECISION_DEFAULT )
		{
	#endif
			lis_matvect(Z,B,X);
			for(i=0;i<n;i++)
			{
				t->value[i] = X->value[i]*d->value[i];
			}
			lis_matvec(W,t,X);
	#ifdef USE_QUAD_PRECISION
		}
		else
		{
			lis_matvect(Z,B,X);
			for(i=0;i<n;i++)
			{
				#ifndef USE_SSE2
					LIS_QUAD_MULD(t->value[i],t->value_lo[i],X->value[i],X->value_lo[i],d->value[i]);
				#else
					LIS_QUAD_MULD_SSE2(t->value[i],t->value_lo[i],X->value[i],X->value_lo[i],d->value[i]);
				#endif
				/* t->value[i] = X->value[i]*d->value[i]; */
			}
			lis_matvec(W,t,X);
		}
	#endif

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
