/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

#define NWORK				9
/************************************************
 * lis_tfqmr_check_params
 * lis_tfqmr_malloc_work
 * lis_tfqmr
 ************************************************/
#undef __FUNC__
#define __FUNC__ "lis_tfqmr_check_params"
int lis_tfqmr_check_params(LIS_SOLVER solver)
{
	LIS_DEBUG_FUNC_IN;
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_tfqmr_malloc_work"
int lis_tfqmr_malloc_work(LIS_SOLVER solver)
{
	LIS_VECTOR	*work;
	int			i,j,worklen,err;

	LIS_DEBUG_FUNC_IN;

	worklen = NWORK;
	work    = (LIS_VECTOR *)lis_malloc( worklen*sizeof(LIS_VECTOR), "lis_tfqmr_malloc_work::work" );
	if( work==NULL )
	{
		LIS_SETERR_MEM(worklen*sizeof(LIS_VECTOR));
		return LIS_ERR_OUT_OF_MEMORY;
	}
	if( solver->precision==LIS_PRECISION_DEFAULT )
	{
		for(i=0;i<worklen;i++)
		{
			err = lis_vector_duplicate(solver->A,&work[i]);
			if( err ) break;
		}
	}
	else
	{
		for(i=0;i<worklen;i++)
		{
			err = lis_vector_duplicateex(LIS_PRECISION_QUAD,solver->A,&work[i]);
			if( err ) break;
		}
	}
	if( i<worklen )
	{
		for(j=0;j<i;j++) lis_vector_destroy(work[j]);
		lis_free(work);
		return err;
	}
	solver->worklen = worklen;
	solver->work    = work;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_tfqmr"
int lis_tfqmr(LIS_SOLVER solver)
{
	LIS_MATRIX A;
	LIS_PRECON M;
	LIS_VECTOR b,x;
	LIS_VECTOR r, rtld, u, p, d, t, t1, q, v;
	LIS_SCALAR tau,rho, rhoold, theta,eta,beta,alpha,w,ww,wold,s,c,etaold,thetaold;
	LIS_SCALAR bnrm2, nrm2;
	LIS_SCALAR tol;
	int iter,maxiter,n,m,output,i;
	double times,ptimes;

	LIS_DEBUG_FUNC_IN;

	A       = solver->A;
	M       = solver->precon;
	b       = solver->b;
	x       = solver->x;
	n       = A->n;
	maxiter = solver->options[LIS_OPTIONS_MAXITER];
	output  = solver->options[LIS_OPTIONS_OUTPUT];
	tol     = solver->params[LIS_PARAMS_RESID-LIS_OPTIONS_LEN];
	ptimes  = 0.0;

	r       = solver->work[0];
	rtld    = solver->work[1];
	u       = solver->work[2];
	p       = solver->work[3];
	d       = solver->work[4];
	t       = solver->work[5];
	t1      = solver->work[6];
	q       = solver->work[7];
	v       = solver->work[8];


	/* Initial Residual */
	if( lis_solver_get_initial_residual(solver,M,r,rtld,&bnrm2) )
	{
		LIS_DEBUG_FUNC_OUT;
		return LIS_SUCCESS;
	}

	/* rtld = r
	   p    = r
	   u    = r
	   d    = 0
	   v    = Ap
	*/
	lis_vector_copy(rtld,r);
	lis_vector_copy(r,p);
	lis_vector_copy(r,u);
	lis_vector_set_all(0,d);
	lis_matvec(A,p,t);
	times = lis_wtime();
	lis_psolve(solver, t, v);
	ptimes += lis_wtime()-times;

	/* rhoold = (r,rtld) */
	lis_vector_dot(r,rtld,&rhoold);
	lis_vector_nrm2(r,&tau);
	wold     = tau;
	theta    = 0.0;
	eta      = 0.0;
	thetaold = 0.0;
	etaold   = 0.0;

	/*	tol = tol * bnrm2; */

	iter=1;
	while( iter<=maxiter )
	{
		/* alpha = rho / (v,rtld) */
		lis_vector_dot(v,rtld,&s);
		/* test breakdown */
		if( s==0.0 )
		{
			solver->retcode   = LIS_BREAKDOWN;
			solver->iter      = iter;
			solver->resid     = nrm2;
			LIS_DEBUG_FUNC_OUT;
			return LIS_BREAKDOWN;
		}
		alpha = rhoold / s;

		/* q = u - alpha*v         */
		/* r = r - alpha*A(u + q)  */
		lis_vector_axpyz(-alpha,v,u,q);
		lis_vector_axpyz(1.0,u,q,t);

		lis_matvec(A,t,t1);
		times = lis_wtime();
		lis_psolve(solver, t1, v);
		ptimes += lis_wtime()-times;
		lis_vector_axpy(-alpha,v,r);

		lis_vector_nrm2(r,&w);
		for(m=0;m<2;m++)
/*		for(m=2*iter-1;m<=2*iter;m++)*/
		{
			if( m==0 )
			{
				ww = sqrt(w*wold);
				lis_vector_xpay(u,theta*theta*eta/alpha,d);
			}
			else
			{
				ww = w;
				lis_vector_xpay(q,theta*theta*eta/alpha,d);
			}
			/*
			if( m%2!=0 )
			{
				ww = w;
				lis_vector_xpay(u,theta*theta*eta/alpha,d);
			}
			else
			{
				ww = sqrt(w*wold);
				lis_vector_xpay(q,theta*theta*eta/alpha,d);
			}
			*/

			theta = ww / tau; 
			c     = 1.0 / sqrt(1.0 + theta*theta);
			eta   = c * c * alpha;
			tau   = tau * theta * c;
			lis_vector_axpy(eta,d,x);


			nrm2 = tau * sqrt(1.0+m) * bnrm2;
			/* convergence check */

			if( m==0 && output )
			{
				if( output & LIS_PRINT_MEM ) solver->residual[iter] = nrm2;
				if( output & LIS_PRINT_OUT && A->my_rank==0 ) printf("iter: %5d  residual = %e\n", iter, nrm2);
			}

			if( tol >= nrm2 )
			{
				solver->retcode    = LIS_SUCCESS;
				solver->iter       = iter;
				solver->resid      = nrm2;
				solver->ptimes     = ptimes;
				LIS_DEBUG_FUNC_OUT;
				return LIS_SUCCESS;
			}
		}

		/* rho  = (r,rtld)              */
		/* beta = rho / rhoold          */
		/* u    = r + beta*q            */
		/* p    = u + beta*(q + beta*p) */
		lis_vector_dot(r,rtld,&rho);
		/* test breakdown */
		if( rho==0.0 )
		{
			solver->retcode   = LIS_BREAKDOWN;
			solver->iter      = iter;
			solver->resid     = nrm2;
			LIS_DEBUG_FUNC_OUT;
			return LIS_BREAKDOWN;
		}
		beta = rho / rhoold;
		lis_vector_axpyz(beta,q,r,u);
/*		lis_vector_axpy(beta,p,q);
		lis_vector_axpyz(beta,q,u,p);*/
		lis_vector_xpay(q,beta,p);
		lis_vector_xpay(u,beta,p);
		lis_matvec(A,p,t1);
		times = lis_wtime();
		lis_psolve(solver, t1, v);
		ptimes += lis_wtime()-times;

		rhoold = rho;
		wold   = w;
		iter++;
	}

	solver->retcode   = LIS_MAXITER;
	solver->iter      = iter;
	solver->resid     = nrm2;
	LIS_DEBUG_FUNC_OUT;
	return LIS_MAXITER;
}

#ifdef USE_QUAD_PRECISION
#undef __FUNC__
#define __FUNC__ "lis_tfqmr_quad"
int lis_tfqmr_quad(LIS_SOLVER solver)
{
	LIS_MATRIX A;
	LIS_PRECON M;
	LIS_VECTOR b,x;
	LIS_VECTOR r, rtld, u, p, d, t, t1, q, v;
	LIS_SCALAR_M tau,rho, rhoold, theta,eta,beta,alpha,w,ww,wold,s,c,etaold,thetaold,one;
	LIS_SCALAR_M nrm2;
	LIS_SCALAR bnrm2, tol;
	int iter,maxiter,n,m,output,i;
	double times,ptimes;

	LIS_DEBUG_FUNC_IN;

	A       = solver->A;
	M       = solver->precon;
	b       = solver->b;
	x       = solver->x;
	n       = A->n;
	maxiter = solver->options[LIS_OPTIONS_MAXITER];
	output  = solver->options[LIS_OPTIONS_OUTPUT];
	tol     = solver->params[LIS_PARAMS_RESID-LIS_OPTIONS_LEN];
	ptimes  = 0.0;

	r       = solver->work[0];
	rtld    = solver->work[1];
	u       = solver->work[2];
	p       = solver->work[3];
	d       = solver->work[4];
	t       = solver->work[5];
	t1      = solver->work[6];
	q       = solver->work[7];
	v       = solver->work[8];

	LIS_QUAD_SCALAR_MALLOC(alpha,0,1);
	LIS_QUAD_SCALAR_MALLOC(beta,1,1);
	LIS_QUAD_SCALAR_MALLOC(rho,2,1);
	LIS_QUAD_SCALAR_MALLOC(rhoold,3,1);
	LIS_QUAD_SCALAR_MALLOC(tau,4,1);
	LIS_QUAD_SCALAR_MALLOC(theta,5,1);
	LIS_QUAD_SCALAR_MALLOC(eta,6,1);
	LIS_QUAD_SCALAR_MALLOC(w,7,1);
	LIS_QUAD_SCALAR_MALLOC(ww,8,1);
	LIS_QUAD_SCALAR_MALLOC(wold,9,1);
	LIS_QUAD_SCALAR_MALLOC(s,10,1);
	LIS_QUAD_SCALAR_MALLOC(c,11,1);
	LIS_QUAD_SCALAR_MALLOC(etaold,12,1);
	LIS_QUAD_SCALAR_MALLOC(thetaold,13,1);
	LIS_QUAD_SCALAR_MALLOC(nrm2,14,1);
	LIS_QUAD_SCALAR_MALLOC(one,15,1);


	/* Initial Residual */
	if( lis_solver_get_initial_residual(solver,M,r,rtld,&bnrm2) )
	{
		LIS_DEBUG_FUNC_OUT;
		return LIS_SUCCESS;
	}

	/* rtld = r
	   p    = r
	   u    = r
	   d    = 0
	   v    = Ap
	*/
	lis_vector_copyex_mm(rtld,r);
	lis_vector_copyex_mm(r,p);
	lis_vector_copyex_mm(r,u);
	lis_vector_set_allex_nm(0.0, d);

	lis_matvec(A,p,t);
	times = lis_wtime();
	lis_psolve(solver, t, v);
	ptimes += lis_wtime()-times;

	/* rhoold = (r,rtld) */
	lis_vector_dotex_mmm(r,rtld,&rhoold);
	lis_vector_nrm2ex_mm(r,&tau);
	wold.hi[0]  = tau.hi[0];
	wold.lo[0]  = tau.lo[0];
	theta.hi[0]    = 0.0;
	theta.lo[0]    = 0.0;
	eta.hi[0]    = 0.0;
	eta.lo[0]    = 0.0;
	thetaold.hi[0]    = 0.0;
	thetaold.lo[0]    = 0.0;
	etaold.hi[0]    = 0.0;
	etaold.lo[0]    = 0.0;
	one.hi[0]    = 1.0;
	one.lo[0]    = 0.0;

	/*	tol = tol * bnrm2; */

	iter=1;
	while( iter<=maxiter )
	{
		/* alpha = rho / (v,rtld) */
		lis_vector_dotex_mmm(v,rtld,&s);
		/* test breakdown */
		if( s.hi[0]==0.0 && s.lo[0]==0.0 )
		{
			solver->retcode   = LIS_BREAKDOWN;
			solver->iter      = iter;
			solver->resid     = nrm2.hi[0];
			LIS_DEBUG_FUNC_OUT;
			return LIS_BREAKDOWN;
		}
		lisqp_div((LIS_QDOUBLE *)alpha.hi,(LIS_QDOUBLE *)rhoold.hi,(LIS_QDOUBLE *)s.hi);

		/* q = u - alpha*v         */
		/* r = r - alpha*A(u + q)  */
		lisqp_minus((LIS_QDOUBLE *)alpha.hi);
		lis_vector_axpyzex_mmmm(alpha,v,u,q);
		lis_vector_axpyzex_mmmm(one,u,q,t);

		lis_matvec(A,t,t1);
		times = lis_wtime();
		lis_psolve(solver, t1, v);
		ptimes += lis_wtime()-times;
		lis_vector_axpyex_mmm(alpha,v,r);

		lis_vector_nrm2ex_mm(r,&w);
		lisqp_minus((LIS_QDOUBLE *)alpha.hi);
		for(m=0;m<2;m++)
/*		for(m=2*iter-1;m<=2*iter;m++)*/
		{
			if( m==0 )
			{
				lisqp_mul((LIS_QDOUBLE *)ww.hi,(LIS_QDOUBLE *)w.hi,(LIS_QDOUBLE *)wold.hi);
				lisqp_sqrt((LIS_QDOUBLE *)ww.hi,(LIS_QDOUBLE *)ww.hi);
				lisqp_sqr((LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)theta.hi);
				lisqp_mul((LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)eta.hi);
				lisqp_div((LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)alpha.hi);
				lis_vector_xpayex_mmm(q,theta,d);
			}
			else
			{
				ww.hi[0] = w.hi[0];
				ww.lo[0] = w.lo[0];
				lisqp_sqr((LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)theta.hi);
				lisqp_mul((LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)eta.hi);
				lisqp_div((LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)alpha.hi);
				lis_vector_xpayex_mmm(u,theta,d);
			}

			lisqp_div((LIS_QDOUBLE *)theta.hi,(LIS_QDOUBLE *)ww.hi,(LIS_QDOUBLE *)tau.hi);

			lisqp_sqr((LIS_QDOUBLE *)c.hi,(LIS_QDOUBLE *)theta.hi);
			lisqp_add((LIS_QDOUBLE *)c.hi,(LIS_QDOUBLE *)c.hi,(LIS_QDOUBLE *)one.hi);
			lisqp_sqrt((LIS_QDOUBLE *)c.hi,(LIS_QDOUBLE *)c.hi);
			lisqp_div((LIS_QDOUBLE *)c.hi,(LIS_QDOUBLE *)one.hi,(LIS_QDOUBLE *)c.hi);

			lisqp_sqr((LIS_QDOUBLE *)eta.hi,(LIS_QDOUBLE *)c.hi);
			lisqp_mul((LIS_QDOUBLE *)eta.hi,(LIS_QDOUBLE *)eta.hi,(LIS_QDOUBLE *)alpha.hi);

			lisqp_mul((LIS_QDOUBLE *)tau.hi,(LIS_QDOUBLE *)tau.hi,(LIS_QDOUBLE *)theta.hi);
			lisqp_mul((LIS_QDOUBLE *)tau.hi,(LIS_QDOUBLE *)tau.hi,(LIS_QDOUBLE *)c.hi);
			lis_vector_axpyex_mmm(eta,d,x);


			lisqp_mul_dd_d((LIS_QDOUBLE *)nrm2.hi,(LIS_QDOUBLE *)tau.hi,sqrt(1.0+m)*bnrm2);
			/* convergence check */

			if( m==0 && output )
			{
				if( output & LIS_PRINT_MEM ) solver->residual[iter] = nrm2.hi[0];
				if( output & LIS_PRINT_OUT && A->my_rank==0 ) printf("iter: %5d  residual = %e %e\n", iter, nrm2.hi[0], nrm2.lo[0]);
			}

			if( (tol > nrm2.hi[0]) || (tol==nrm2.hi[0] && nrm2.lo[0]<=0.0) )
			{
				solver->retcode    = LIS_SUCCESS;
				solver->iter       = iter;
				solver->resid      = nrm2.hi[0];
				solver->ptimes     = ptimes;
				LIS_DEBUG_FUNC_OUT;
				return LIS_SUCCESS;
			}
		}

		/* rho  = (r,rtld)              */
		/* beta = rho / rhoold          */
		/* u    = r + beta*q            */
		/* p    = u + beta*(q + beta*p) */
		lis_vector_dotex_mmm(r,rtld,&rho);
		/* test breakdown */
		if( rho.hi[0]==0.0 && rho.lo[0]==0.0 )
		{
			solver->retcode   = LIS_BREAKDOWN;
			solver->iter      = iter;
			solver->resid     = nrm2.hi[0];
			LIS_DEBUG_FUNC_OUT;
			return LIS_BREAKDOWN;
		}
		lisqp_div((LIS_QDOUBLE *)beta.hi,(LIS_QDOUBLE *)rho.hi,(LIS_QDOUBLE *)rhoold.hi);
		lis_vector_axpyzex_mmmm(beta,q,r,u);
		lis_vector_xpayex_mmm(q,beta,p);
		lis_vector_xpayex_mmm(u,beta,p);
		lis_matvec(A,p,t1);
		times = lis_wtime();
		lis_psolve(solver, t1, v);
		ptimes += lis_wtime()-times;

		rhoold.hi[0] = rho.hi[0];
		rhoold.lo[0] = rho.lo[0];
		wold.hi[0] = w.hi[0];
		wold.lo[0] = w.lo[0];
		iter++;
	}

	solver->retcode   = LIS_MAXITER;
	solver->iter      = iter;
	solver->resid     = nrm2.hi[0];
	LIS_DEBUG_FUNC_OUT;
	return LIS_MAXITER;
}
#endif
